#############################################################################
##
#W    subgroupcomplexes.gd    share package 'homology'                 Welker
##
## Partially ordered sets and simplicial complexes arising in the study    
## of the structure of the subgroup-lattice and parts of it of a finite group.

#==================================================
# Posets from groups 
#==================================================

LocalPosetOfConjugacyClasses := function (l)
  return List(MinimalSupergroupsLattice(l), c -> 
              DuplicateFreeList(List(c,e -> e[1]))); 
end;

InstallGlobalFunction (PosetOfConjugacyClasses, function(g)
  return LocalPosetOfConjugacyClasses(LatticeSubgroups(g));
end);


ConjugacyClassesToSubgroups := function (selected_conjugacy_classes, gap_lattice) 
  local ret_lattice, small_conjugacy_class, large_conjugacy_class, 
        selected_group_list, small_group, large_group, 
        number_of_conjugacy_classes, list_selected, index_list, i;
  
  number_of_conjugacy_classes := Number(ConjugacyClassesSubgroups(gap_lattice));
 
  for i in [1..number_of_conjugacy_classes-1] do
    if (selected_conjugacy_classes[i] = []) then
      selected_conjugacy_classes[i]:=[number_of_conjugacy_classes];
    fi;
  od;
 
  index_list := [1];
  selected_group_list := [1];
  for i in [2..number_of_conjugacy_classes] do
    if not (selected_conjugacy_classes[i-1]  = []) then
      Append(index_list,[index_list[i-1]+Number(ConjugacyClassesSubgroups(gap_lattice)[i-1])]);
      Append(selected_group_list,[i-1]);
    else
      Append(index_list,[index_list[i-1]]);
    fi;
  od;


  ret_lattice := []; 
  for small_conjugacy_class in Difference(selected_group_list,[number_of_conjugacy_classes]) do
    for small_group in 
          [1..Number(ConjugacyClassesSubgroups(gap_lattice)[small_conjugacy_class])] do
      ret_lattice := Concatenation(ret_lattice,[[]]);
      for large_conjugacy_class in selected_conjugacy_classes[small_conjugacy_class] do
        if (large_conjugacy_class = number_of_conjugacy_classes) then
          ret_lattice[index_list[small_conjugacy_class]+small_group-1] := 
                 Concatenation(ret_lattice[index_list[small_conjugacy_class]+small_group-1],
                               [index_list[large_conjugacy_class]]);
        else
          for large_group in 
                 [1..Number(ConjugacyClassesSubgroups(gap_lattice)[large_conjugacy_class])] do
            if (IsSubgroup(
              ConjugacyClassesSubgroups(gap_lattice)[large_conjugacy_class][large_group],
              ConjugacyClassesSubgroups(gap_lattice)[small_conjugacy_class][small_group])) then
              ret_lattice[index_list[small_conjugacy_class]+small_group-1] :=
              Concatenation(ret_lattice[index_list[small_conjugacy_class]+small_group-1],
                                  [index_list[large_conjugacy_class]+large_group-1]);
            fi;
          od;
        fi;
      od;
    od;
  od;

  return Concatenation(ret_lattice,[[]]);
end;


InstallGlobalFunction (LatticeOfSubgroups, function(g)
  local gap_lattice, poset_conjugacy_classes,ret_lattice;

  gap_lattice := LatticeSubgroups(g);
  poset_conjugacy_classes := LocalPosetOfConjugacyClasses(gap_lattice);
  return ConjugacyClassesToSubgroups(poset_conjugacy_classes,gap_lattice);
end);

InstallGlobalFunction (IsElementaryAbelianPGroup, function (g,p)
  local order_g, primes_order_g, return_value;

  if not IsPrime (p) then
    Info (InfoWarning, 1, "IsElementaryAbelianPGroup: the second argument should be a prime number");
    return fail;
  fi;

  order_g := Order(g);

  return_value:=false;

  if order_g = 1 then 
    return_value := true;
  else
    primes_order_g := PrimePowersInt(order_g);
    if ((Size(primes_order_g) = 2) and (primes_order_g[1] = p)) then
      if IsElementaryAbelian(g) then
        return_value := true;
      fi;
    fi;
  fi;
  
  return return_value;
end);


IsInQuillenPoset := function(g,gap_lattice, h_conjugacy_class , 
                             p, number_of_conjugacy_classes)
  local return_value;

  return_value := false;
  if (h_conjugacy_class = number_of_conjugacy_classes) then
    return_value := IsElementaryAbelianPGroup(g,p);
  else
    return_value := IsElementaryAbelianPGroup(
                       ConjugacyClassesSubgroups(gap_lattice)
                                       [h_conjugacy_class][1],p);
  fi;
  return return_value;
end;


InstallGlobalFunction (QuillenPoset, function (g,p)
  local lattice_of_subgroups, number_of_subgroups,
        QuillenPosetNonElAb;

  QuillenPosetNonElAb := function (g,p)
  local poset_conjugacy_classes, quillen_conjugacy_poset, gap_lattice,
        number_of_conjugacy_classes;

    gap_lattice := LatticeSubgroups(g);
    poset_conjugacy_classes := List(MinimalSupergroupsLattice(gap_lattice), c ->
                                 DuplicateFreeList(List(c,e -> e[1])));
    number_of_conjugacy_classes := Number(ConjugacyClassesSubgroups(gap_lattice));
 
    quillen_conjugacy_poset := List(poset_conjugacy_classes, succ -> 
                                 Filtered(succ, grp -> 
                                   IsInQuillenPoset 
                                     (g,gap_lattice,grp,p,number_of_conjugacy_classes)));
    return ConjugacyClassesToSubgroups(quillen_conjugacy_poset, gap_lattice);
  end;

  if not IsPrime (p) then
    Info (InfoWarning, 1, "QuillenPoset: the second argument should be a prime number");
    return fail;
  fi;

  if IsElementaryAbelianPGroup(g,p) then
    lattice_of_subgroups := LatticeOfSubgroups(g);
    number_of_subgroups  := Number(lattice_of_subgroups);
    lattice_of_subgroups[number_of_subgroups] := [number_of_subgroups+1];
    return Concatenation(lattice_of_subgroups,[[]]);
  else
    return QuillenPosetNonElAb (g,p);
  fi;
end); 


InstallGlobalFunction (PosetOfCosets, function (g)
  local gap_lattice, lattice_of_subgroups, number_of_subgroups, 
        index_list, cosets_list, ret_poset, small_group, small_coset,
        large_group, large_coset, i, j, l, number_of_conjugacy_classes,
        number_of_cosets, large_coset_list, small_coset_list,
        conjugacy_class,small_index;
 
  gap_lattice := LatticeSubgroups(g);
  lattice_of_subgroups := ConjugacyClassesToSubgroups(
                             LocalPosetOfConjugacyClasses(gap_lattice),
                             gap_lattice);
  
  number_of_conjugacy_classes := Number(ConjugacyClassesSubgroups(gap_lattice));
  number_of_subgroups := Number(lattice_of_subgroups);
 
  index_list := [2];
  cosets_list := [];

  l := 1;
  for i in [1..number_of_conjugacy_classes] do
    if (i = number_of_conjugacy_classes) then 
      Append(cosets_list, [RightCosets(g,g)]);
      Append(index_list,[index_list[l]+1]);
      l := l+1;
    else
      conjugacy_class := ConjugacyClassesSubgroups(gap_lattice)[i];
      Append(cosets_list, 
             [RightCosets(g,conjugacy_class[1])]);
      number_of_cosets := Number(cosets_list[l]);
      Append(index_list,[index_list[l]+number_of_cosets]);
      l := l+1;
      for j in [2..Number(conjugacy_class)] do
        Append(cosets_list, 
               [RightCosets(g,conjugacy_class[j])]);
        Append(index_list,[index_list[l]+number_of_cosets]);
        l := l+1;
      od;
    fi;
  od;

  ret_poset := [[2..index_list[2]-1]]; 
  for small_group in [1..number_of_subgroups-1] do
    small_coset_list := cosets_list[small_group];
    for small_coset in [1..Number(small_coset_list)] do
      small_index := index_list[small_group]+small_coset-1;
      ret_poset := Concatenation(ret_poset,[[]]);
      for large_group in lattice_of_subgroups[small_group] do
        if (large_group = number_of_subgroups) then
          ret_poset[small_index] := 
                 Concatenation(ret_poset[small_index],
                               [index_list[large_group]]);
        else
          large_coset_list := cosets_list[large_group];
          for large_coset in 
                 [1..Number(large_coset_list)] do
            if IsSubset(large_coset_list[large_coset],
                        small_coset_list[small_coset]) then
              ret_poset[small_index] :=
                    Concatenation(ret_poset[small_index],
                                  [index_list[large_group]+large_coset-1]);
            fi;
          od;
        fi;
      od;
    od;
  od;

  return Concatenation(ret_poset,[[]]);
end);


