/****************************************************************************
**
*W  commentator.h          Homology source                  DHS 
**
*H  @(#)$Id: commentator.h,v 0.0 1999/08
**
*Y  Copyright (C)  1999,  Us guys
**
**  This file implements the C++ interface to commentators (for 
**  providing runtime commentary to the user)
*/

#ifndef __commentary_H
#define __commentary_H
#include <deque.h>
#include <stack.h>
#include <givtimer.h>
//#include <limits.h>

// =============================================
// 
// Message classes definitions
// 
// =============================================

// Min and Max
// initializing commentator with PRINT_EVERYTHING, everything should be printed.
// initializing commentator with PRINT_NOTHING, nothing should be printed.
#define PRINT_EVERYTHING 100000 
#define PRINT_NOTHING 0

// Msg Class
#define TIMING_MEASURE          1
#define TIMING_ESTIMATE         2
#define PARTIAL_RESULT          3
#define INTERNAL_WARNING        4
#define INTERNAL_ERROR          5
#define INTERNAL_DESCRIPTION    6

// Msg Level
#define LVL_ALWAYS      -1
#define LVL_IMP         0
#define LVL_NORMAL      1
#define LVL_UNIMP       2
#define LVL_BLABLA      10
//#define LVL_NEVER       INT_MAX
#define LVL_NEVER       (PRINT_EVERYTHING+1)


// n th root of a double 
#define _NROOT_PRECISION_ 0.0001
double nroot(double a, long r, double precision);
long isnpower(long& l, long a) ;


// =============================================
// 
// Commentator class
// 
// =============================================
class Commentator {
public: 
    Commentator() : gapTimingLevel(0), gapHomologyLevel(0), indentationStep(0), estimatesDepth(0) {};
        // When copying, Timers and Estimators are not needed.
    Commentator(const Commentator& C) : gapTimingLevel(C.gapTimingLevel), gapHomologyLevel(C.gapHomologyLevel), indentationStep(C.indentationStep)
			, IDs(C.IDs), Timers(stack<Timer>())
			, estimatesDepth(C.estimatesDepth)
			, Estimates(stack<Estimator>()) {};

        // Sets user printing levels
        // tab is the indentation unit
        // depth is the number of steps stored for time estimates (should be greater than 2)
    Commentator(long timinglvl, long homologylvl, long tab , long depth);

    void start(const char* id, const char* msg, long msglevel, long msgclass);
    void stop(const char* msg, long msglevel, long msgclass);
    
    void progress(const char* msg, long msglevel, long k, long n);

    void report(const char* msg, long msglevel, long msgclass) const;
    long printed(long msglevel, long msgclass) const;
private:
        // Internal sub routines
    long intern_printed(char& information, long msglevel, long msgclass) const;
    long header(long msglevel, long msgclass) const ;
    void printheader(char information) const ;

        // Printing decision is made from those
    long gapTimingLevel;
    long gapHomologyLevel;

        // Activity id
    long indentationStep;
    stack<const char*> IDs;

        // Activity Timer
    stack<Timer> Timers;

        // Activity Estimator
    long estimatesDepth;

    class StepsAndTime {
    public:
        StepsAndTime(long k, double t) : _time(t), _steps(k) {}
        long getsteps() { return _steps; }
        double gettime() { return _time; }
        
    private:
        double _time;
        long _steps;
    };
        
    typedef deque< StepsAndTime > Estimator;
    stack<Estimator> Estimates;
};

// =============================================
// 
// wrappers for use by Pascal (& C) code.
// 
// =============================================
extern "C" Commentator* initializeCommentator(long timing, long homology);

extern "C" void startActivity(Commentator& C, const char* id, const char* msg, long msglevel, long msgclass);

extern "C" void stopActivity(Commentator& C, const char* msg, long msglevel, long msgclass);

extern "C" void activityReport(const Commentator& C, const char* msg, long msglevel, long msgclass);

extern "C" void progressReport(Commentator& C, const char* msg, long msglevel, long k, long n);

extern "C" long isPrinted(const Commentator& C, long msglevel, long msgclass);

#include "commentator.inl"

#endif
