/****************************************************************************
**
*W  smith_form.C          Homology source                  Dumas**
*H  @(#)$Id: smith_form.C,v 0.0 1999/08
**
*Y  Copyright (C)  1999,  Us guys
*T Time-stamp: <31 Jan 03 16:01:20 Jean-Guillaume.Dumas@imag.fr> 
**
*/
#include <stdio.h> 


#include "matrix.h"

#include "commentator.h"
Commentator GlobalCommentator;


#include "givinteger.h"
#include "givintfactor.h"

#define MINERRORLEVEL 0.1

// -------------------------------------------------
// Finite Fields
// -------------------------------------------------
#include "givzpz.h"
typedef ZpzDom<Std64>               GFqDomain;
typedef GFqDomain::Residu_t        Residu;
#define MINLOWERSIZE 1073741824ULL
#define MAXPRIMESIZE 2147483648ULL
#define NUMBEROFPRIMES 50697922ULL
#define WIED_EARLYTERM_THRESHOLD 20
#define RNS_EARLYTERM_THRESHOLD 10


// -------------------------------------------------
// Sparse structure and Methods
// -------------------------------------------------
#include "LinBox/lin_spv_bb.h" 
#include "LinBox/lin_cassini.h"
#include "LinBox/lin_invariants.h"
#include "LinBox/lin_valence.C"
#include "LinBox/lin_wiedemann_rank.C"
#include "LinBox/lin_gauss_smith.C"

#define WITHINTEGERRANK 1
#define WITHOUTINTEGERRANK 0




void MatrixSmithFormValenceCPP (Commentator & Comm,
                                long ni, long nj,
                                char * File_Name,
                                int Method, int Mode, Integer& R,
                                Invariants<unsigned long long>& Inv, 
				char * numerator, char * denominator) {
    startActivity(GlobalCommentator, "SNFV", "Smith Normal Form of Integer Matrix using Valence algorithm", LVL_NORMAL, INTERNAL_DESCRIPTION ); 
    activityReport(GlobalCommentator, "Loading the matrix ...", LVL_BLABLA, INTERNAL_DESCRIPTION);



        // Here I make a copy of the Commentator
    GlobalCommentator = Comm;


    double error = double(Integer(numerator))/double(Integer(denominator));
//     cerr << Integer(numerator) << "/" << Integer(denominator) << " = " << error << endl;
    double wrong = 0, int_wrong;

    Inv = Invariants<unsigned long long>(ni, nj);


    Integer Valence, p1; 
    bool iswithinteger;
    unsigned long rank;
    
    char * UT = new char[270]; 
  
    switch (Mode) {
        case SMITHFORM:
        case INTEGRANK: 
        {
            startActivity(GlobalCommentator, "Gersh", "Cassini Ovals method, to create a bound on the coefficients of A^t . A", LVL_NORMAL, INTERNAL_DESCRIPTION ); 
            Cassini<double>::Type_t OCB;
            { Cassini<double>(File_Name).operator() (OCB); }
            sprintf(UT,"Cassini eigen value bound : %f", OCB); 
            stopActivity(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT ); 

            { seq_Integer_valence(File_Name, Valence, rank, OCB, wrong, error); }

            iswithinteger = WITHINTEGERRANK;
            
            break;
        }
        case TORSIONRANK:
        {
            Valence = R;
            iswithinteger = WITHOUTINTEGERRANK;
        }
    }
 
    if (Method == hmValenceBB) {
        switch (Mode) {
            case SMITHFORM:
            case TORSIONRANK:
                wiedemann_rank_N_torsion(Inv, File_Name, Valence); 
                break;
            case INTEGRANK:
                wiedemann_Integer_rank(rank, p1, File_Name, Valence, wrong, error); 
                break; 
            case PRIMERANK:
                double rank_wrong; 
                wiedemann_rank(rank, File_Name, (unsigned long)R, rank_wrong); 
                break;
            case PRIMEPOWERRANK:
                unsigned long expo; Integer prime;
                IntFactorDom<> IPD;
                if (! IPD.isprimepower(prime, expo, R) ) {
                    string sr = (string)R, sp = (string)prime;
                    sprintf(UT,"%s is not a prime power, assuming %s^%ld instead.", sr.c_str(), sp.c_str(), expo);
                    activityReport(GlobalCommentator, UT, LVL_IMP, INTERNAL_DESCRIPTION);
                }
                if (expo > 1) {
                    activityReport(GlobalCommentator, "Sorry, prime power rank via Wiedemann is NOT YET AVAILABLE, computing prime rank instead.",LVL_NORMAL, INTERNAL_ERROR);
                }
                
                wiedemann_rank(rank, File_Name, (unsigned long)prime, rank_wrong); 
       }
    } else {
        switch (Mode) {
            case SMITHFORM:
            case TORSIONRANK:
                gauss_rank_N_torsion(Inv, File_Name, Valence, iswithinteger); 
                break;
            case INTEGRANK:
                gauss_Integer_rank(rank, p1, File_Name, Valence, wrong, error); 
                break; 
            case PRIMERANK:
                one_gauss(rank, File_Name, (long)R);  
                break; 
            case CARACTERISTICROWS:
                one_permuts(Inv, rank, File_Name, (long)R);  
                break; 
            case PRIMEPOWERRANK:
                unsigned long expo; Integer prime;
                IntFactorDom<> IPD;
                if (! IPD.isprimepower(prime, expo, R) ) {
                    string sr = (string)R, sp = (string)prime;
                    sprintf(UT,"%s is not a prime power, assuming %s^%ld instead.", sr.c_str(), sp.c_str(), expo);
                    activityReport(GlobalCommentator, UT, LVL_IMP, INTERNAL_DESCRIPTION);
                    IPD.pow(R, prime, expo);
                }
                
                one_pp_gauss(rank, File_Name, (long)R, (long)prime);
        }
    }

    switch (Mode) {
        case PRIMEPOWERRANK:
        case PRIMERANK:
            Inv.PushInvariant( (long)R, rank);
            break;
        case INTEGRANK:
            if (wrong == 1)
                sprintf(UT,"Rank %ld mod %ld is the Integer rank with unknown probability ", rank, (long)p1);
            else
                sprintf(UT,"Rank %ld mod %ld is the Integer rank with probability at least 1 - %e", rank, (long)p1, wrong );
            activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
            Inv.PushInvariant( 1, rank);
    }
    

    delete [] UT;
}
